@@ -394,6 +394,11 @@ function recreateProController() {
     socket.emit('create_pro_controller');
 }
 
+function restartController() {
+    shutdownController();
+    setTimeout(recreateProController, 2000);
+}
+
 function checkForLoad() {
     if (STATE[NXBT_CONTROLLER_INDEX]) {
         controller_state = STATE[NXBT_CONTROLLER_INDEX].state
@@ -463,6 +468,22 @@ function changeInput(evt) {
     }
 }
 
+function changeFrequency(evt) {
+    let newFrequency = evt.target.value;
+    
+    if (newFrequency === "RAF") {
+        useRAF = true;
+    } else {
+        newFrequency = Number(newFrequency);
+        if (!isNaN(newFrequency)) {
+            useRAF = false;
+            frequency = (1/newFrequency) * 1000;
+        } else {
+            console.log("New frequency is not a number");
+        }
+    }
+}
+
 const LOADER_ANIMATION_FRAMES = [0,1,2,3,3,2,1,0];
 let loaderFrame = 1;
 let highlightedBlock = false;
@@ -548,6 +569,7 @@ function updateGamepadDisplay() {
 
 let timeOld = false;
 let frequency = (1/120) * 1000;
+let useRAF = true;
 function eventLoop() {
     // Update x/y ratio for the sticks based on
     // pressed buttons if we're using a keyboard
@@ -601,21 +623,23 @@ function eventLoop() {
 
     updateGamepadDisplay()
 
-    // if (!timeOld) {
-    //     timeOld = performance.now();
-    // }
-    // timeNew = performance.now();
-    // delta = timeNew - timeOld;
-    // diff = delta - frequency;
-
-    // if (diff > 0) {
-    //     setTimeout(eventLoop, frequency - diff);
-    // } else {
-    //     setTimeout(eventLoop, frequency);
-    // }
-    // timeOld = timeNew;
-
-    requestAnimationFrame(eventLoop);
+    if (useRAF) {
+        requestAnimationFrame(eventLoop);
+    } else {
+        if (!timeOld) {
+            timeOld = performance.now();
+        }
+        timeNew = performance.now();
+        delta = timeNew - timeOld;
+        diff = delta - frequency;
+    
+        if (diff > 0) {
+            setTimeout(eventLoop, frequency - diff);
+        } else {
+            setTimeout(eventLoop, frequency);
+        }
+        timeOld = timeNew;
+    }
 }
 
 function sendMacro() {
